const _ = require('lodash');
const async = require('async');
const DEBUG = require('debug');

const Error = require('./error');
const utils = require('./utils');
const config = require('./config');
const handlers = require('./request');
const interfaces = require('./interfaces');
const sessionManager = require('./session');

class FacePlatform {
    constructor (option) {
        option = option || {};
        if (!config.versions.includes(option.version)) {
            throw Error(80001);
        }

        option.address = option.address || config.default_address;
        this.address = option.address.startsWith('http') ? option.address : 'http://' + option.address;
        this.version = option.version;
        this.auth = option.auth || config.default_auth;
        this.sessionManagerId  = utils.md5(this.address, this.version, this.auth.name);
        this.request_timeout = option.request_timeout || config.request_timeout;
        this.chains = [];
        this.debug = DEBUG("faceplatform");
        this.utils = utils;

        const fp = new Proxy(this, {
            get: (target, key) => {
                if (interfaces[key]) {
                    return (...args) => {
                        this.chains.push(handlers.requestHandler.bind(target, key, ...args));
                        return fp;
                    };
                } else {
                    return target[key];
                }
            }
        });

        return fp;
    }

    exec (next) {
        this._finally = next;
        this.debug('FacePlatform address: ' + this.address + ' with version: ' + this.version);
        this.debug('start revoking');

        let p = sessionManager.getSession(this.sessionManagerId);
        p = p.then(sessionId => {
            if (sessionId) {
                this.session_id = sessionId;
                return Promise.resolve();
            } else {
                sessionManager.setRefreshing(this.sessionManagerId, true);
                return handlers.login.apply(this).then(resp => {
                    sessionManager.setSession(this.sessionManagerId, resp.session_id);
                    sessionManager.setRefreshing(this.sessionManagerId, false);
                    this.session_id = resp.session_id;
                });
            }
        });


        while (this.chains.length) {
            const handler = this.chains.shift();
            const wrap_f = () => {
                if (this.to_break) {
                    return Promise.reject(Error(80201));
                }

                return new Promise((resolve, reject) => {
                    sessionManager.tick(this.sessionManagerId);
                    let result = handler();
                    if (result instanceof FacePlatform) {
                        // somehow the chains auto breaks if an instance of FP is returned
                        this.debug('this chain function returns an instance of FacePlatform, which is not allowed, undefined is returned instead.');
                        result = undefined;
                    }
                    return resolve(result);
                });
            };
            p = p.then(wrap_f);
        }
        p.catch(err => {
            if (config.forgivable_errors.includes(err.error_code)) {
                this.debug('forgivable error: %s', err.message);
            } else {
                this.throw(err);
            }
        })
        .then(() => {
            this._done();
        });
    }

    _done () {
        this.to_break = false;
        this.chains = [];
        const _f = this._finally;
        delete this._finally;
        this.debug('_done revoking, execute final function');
        _.isFunction(_f) && _f();
    }

    break () {
        this.debug('manually stop chaining here');
        this.to_break = true;
    }

    then (next) {
        this.chains.push(next.bind(this));
        return this;
    }

    throw (err) {
        this.debug(err);
        if (_.isFunction(this._catch)) {
            this._catch(err);
            delete this._catch;
        } else {
            throw err;
        }
    }

    catch (next) {
        this._catch = next;
        return this;
    }

    w (func) {
        const _w_f = () => {
            return func();
        }
        _w_f.w = true;
        return _w_f;
    }
};

module.exports = FacePlatform;
